/**
 * Copyright (c) 2012 - 2016 Data In Motion and others.
 * All rights reserved. 
 * 
 * This program and the accompanying materials are made available under the terms of the 
 * Eclipse Public License v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     Data In Motion - initial API and implementation
 */
package org.gecko.core.tests;

import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.atomic.AtomicInteger;

import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;
import org.osgi.util.promise.Deferred;
import org.osgi.util.promise.Promise;
import org.osgi.util.promise.PromiseFactory;
import org.osgi.util.tracker.ServiceTrackerCustomizer;

/**
 * Customizer that tracks the service providers
 * Each call of add, modify or remove will be counted. Further a {@link CountDownLatch} can be given,
 * to ensure to wait till the service is available
 * @author Mark Hoffmann
 * @since 16.06.2016
 */
public class ServiceCheckerCustomizer<S, T>
		implements ServiceTrackerCustomizer<S, T> {
	
	private final Map<ServiceReference<S>, T> serviceMap = new ConcurrentHashMap<>();
	private final AtomicInteger addCount = new AtomicInteger();
	private final AtomicInteger modifyCount = new AtomicInteger();
	private final AtomicInteger removeCount = new AtomicInteger();
	private final AtomicInteger addExpectation = new AtomicInteger();
	private final AtomicInteger modifyExpectation = new AtomicInteger();
	private final AtomicInteger removeExpectation = new AtomicInteger();
	private Deferred<Boolean> addDef;
	private Deferred<Boolean> modifyDef;
	private Deferred<Boolean> removeDef;
	private final BundleContext context;
	private final PromiseFactory pf;
	
	public ServiceCheckerCustomizer(BundleContext context, PromiseFactory pf) {
		this.context = context;
		this.pf = pf;
	}
	
	/**
	 * Returns how often add was called
	 * @return how often add was called
	 */
	public int getAddCount() {
		return addCount.get();
	}
	
	/**
	 * Returns how often remove was called
	 * @return how often remove was called
	 */
	public int getRemoveCount() {
		return removeCount.get();
	}
	
	/**
	 * Returns how often modify was called
	 * @return how often modify was called
	 */
	public int getModifyCount() {
		return modifyCount.get();
	}
	
	/**
	 * Returns the currentReference.
	 * @return the currentReference
	 */
	public Set<ServiceReference<S>> getReferences() {
		synchronized (serviceMap) {
			return Collections.unmodifiableSet(serviceMap.keySet());
		}
	}
	
	/**
	 * Returns the currentService.
	 * @return the currentService
	 */
	public Set<T> getServices() {
		synchronized (serviceMap) {
			return Collections.unmodifiableSet(new HashSet<>(serviceMap.values()));
		}
	}
	
	public T getService(ServiceReference<S> reference) {
		synchronized (serviceMap) {
			return serviceMap.get(reference);
		}
	}
	
	/* 
	 * (non-Javadoc)
	 * @see org.osgi.util.tracker.ServiceTrackerCustomizer#addingService(org.osgi.framework.ServiceReference)
	 */
	@SuppressWarnings("unchecked")
	@Override
	public T addingService(ServiceReference<S> reference) {
		T service  = null;
		try {
			service  = (T)context.getService(reference);
			synchronized (serviceMap) {
				if (service != null && reference != null) {
					serviceMap.put(reference, service);
					checkAndResolve(addDef, addExpectation.get(), addCount.incrementAndGet(), true);
				}
			}
		} catch (Exception e) {
		}
		return service;
	}

	/* 
	 * (non-Javadoc)
	 * @see org.osgi.util.tracker.ServiceTrackerCustomizer#modifiedService(org.osgi.framework.ServiceReference, java.lang.Object)
	 */
	@Override
	public void modifiedService(ServiceReference<S> reference, T service) {
		synchronized (serviceMap) {
			if (service != null && reference != null) {
				serviceMap.put(reference, service);
			}
		}
		checkAndResolve(modifyDef, modifyExpectation.get(), modifyCount.incrementAndGet(), true);
	}

	/* 
	 * (non-Javadoc)
	 * @see org.osgi.util.tracker.ServiceTrackerCustomizer#removedService(org.osgi.framework.ServiceReference, java.lang.Object)
	 */
	@Override
	public void removedService(ServiceReference<S> reference, T service) {
		synchronized (serviceMap) {
			serviceMap.remove(reference);
		}
		checkAndResolve(removeDef, removeExpectation.get(), removeCount.incrementAndGet(), true);
	}
	
	public void dispose() {
		addCount.set(0);
		removeCount.set(0);
		modifyCount.set(0);
		addExpectation.set(0);
		modifyExpectation.set(0);
		removeExpectation.set(0);
		addDef = null;
		modifyDef = null;
		removeDef = null;
		serviceMap.clear();
	}
	
	public Promise<Boolean> resolveShutdown(int removeOffset) {
		int toRemove = addCount.get();
		if (removeOffset > 0) {
			toRemove -= removeOffset;
		}
		toRemove = toRemove > 0 ? toRemove : 0;
		// Nothing to wait for
		if (toRemove == 0) {
			return pf.resolved(Boolean.TRUE);
		}
		return resolveRemove(toRemove, 0);
	}
	
	public Promise<Boolean> resolveCreate(int expected, int milliseconds) {
		setCreateDeferred(null, milliseconds);
		addExpectation.set(expected);
		return setCreateDeferred(pf.deferred(), milliseconds);
	}
	
	public Promise<Boolean> resolveModify(int expected, int milliseconds) {
		setModifyDeferred(null, milliseconds);
		modifyExpectation.set(expected);
		return setModifyDeferred(pf.deferred(), milliseconds);
	}
	
	public Promise<Boolean> resolveRemove(int expected, int milliseconds) {
		setRemoveDeferred(null, milliseconds);
		removeExpectation.set(expected);
		return setRemoveDeferred(pf.deferred(), milliseconds);
	}
	
	private Promise<Boolean> setCreateDeferred(Deferred<Boolean> d, int milliseconds) {
		if (d == null) {
			addDef = null;
			return null;
		} else {
			addDef = d;
			checkAndResolve(addDef, addExpectation.get(), addCount.get(), milliseconds > 0);
			Promise<Boolean> p = addDef.getPromise();
			return getRecoverablePromise(p, addExpectation, addCount, milliseconds);
		}
	}

	private Promise<Boolean> setModifyDeferred(Deferred<Boolean> d, int milliseconds) {
		if (d == null) {
			modifyDef = null;
			return null;
		} else {
			modifyDef = d;
			checkAndResolve(modifyDef, modifyExpectation.get(), modifyCount.get(), milliseconds > 0);
			Promise<Boolean> p = modifyDef.getPromise();
			return getRecoverablePromise(p, modifyExpectation, modifyCount, milliseconds);
		}
	}

	private Promise<Boolean> setRemoveDeferred(Deferred<Boolean> d, int milliseconds) {
		if (d == null) {
			removeDef = null;
			return null;
		} else {
			removeDef = d;
			checkAndResolve(removeDef, removeExpectation.get(), removeCount.get(), milliseconds > 0);
			Promise<Boolean> p = removeDef.getPromise();
			return getRecoverablePromise(p, removeExpectation, removeCount, milliseconds);
		}
	}
	
	private Promise<Boolean> getRecoverablePromise(Promise<Boolean> p, AtomicInteger expectation, AtomicInteger value, int milliseconds) {
		final int e = expectation.get();
		final int v = value.get();
		return milliseconds > 0 ? p.timeout(milliseconds).recoverWith(f->recover(f, e, v)) : p;
	}
	
	private Promise<Boolean> recover(Promise<?> f, int expected, int value) throws Exception {
		return pf.resolved(Boolean.valueOf(expected == value));
	}

	private void checkAndResolve(Deferred<Boolean> deferred, int expected, int value, boolean wait) {
		if (deferred == null || 
				deferred.getPromise().isDone()) {
			return;
		}
		Boolean result = null;
		if (value > expected) {
			result = Boolean.FALSE;
		} else if (!wait) {
			result = Boolean.valueOf(expected == value);
		} else if (expected > 0 && expected == value) {
			result = Boolean.TRUE;
		}
		if (deferred != null && result != null) {
			deferred.resolve(result);
		}
	}

}